//***********************************************************************//
//*        ELEMENT.H --- header file for ELEMENT.CPP                    *//
//*                                                                     *//
//*Copyright (C) Michael Ortega-Binderberger (miki@ics.uci.edu)         *//
//*Univ of California,Irvine Department of Information&Computer Science *//
//*                                                                     *//
//*This program is free software; you can redistribute itand/or         *//
//*modify it under the terms of the GNU General Public License          *//
//*as published by the Free Software Foundation; either version 2       *//
//*of the License, or (at your option) any later version.               *//
//*                                                                     *//
//*This program is distributed in the hope that it will be useful,      *//
//*but WITHOUT ANY WARRANTY; without even the implied warranty of       *//
//*MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *//
//*GNU General Public License for more details.                         *//
//*                                                                     *//
//*You should have received a copy of the GNU General Public License    *//
//*along with this program; if not, write to the Free Software          *//
//*Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            *//
//*02111-1307, USA.                                                     *//
//*                                                                     *//
//*                                                                     *//
//*This code was licensed under the GNU public license. For our         *//
//*modifications,the Regents of the University of California demand the *//
//*following terms: (where a conflic occurs, the lesser restrictive term*//
//*applies)                                                             *//
//*                                                                     *//
//*Copyright 1996-99 The Regents of the University of California.       *//
//*All Rights Reserved.                                                 *//
//*Permission to use, copy, modify, and distribute this software        *//
//*and its documentation for educational, research and non-profit       *//
//*purposes, without fee, and without a written agreement is hereby     *//
//*granted, provided that the above copyright notice, this paragraph    *//
//*and the following three paragraphs appear in all copies.             *//
//*Permission to incorporate this software into commercial products     *//
//*may be obtained by contacting the University of California.          *//
//*Chen Li Department of Information and Computer Science (ICS)         *//
//*University of California Irvine, California 92697-3425               *//
//*Phone: 949-824-9470.                                                 *//
//*                                                                     *//
//*This software program and documentation are copyrighted by           *//
//*The Regents of the University of California. The software program    *//
//*and documentation are supplied "as is", without any accompanying     *//
//*services from The Regents. The Regents do not warrant that the       *//
//*operation of the program will be uninterrupted or error-free.        *//
//*The end-user understands that the program was developed for          *//
//*research purposes and is advised not to rely exclusively on          *//
//*the program for any reason.                                          *//
//*                                                                     *//
//*IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO          *//
//*ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR              *//
//*CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT           *//
//*OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE       *//
//*UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF      *//
//*SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS     *//
//*ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           *//
//*WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  *//
//*THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE      *//
//*UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,  *//
//*SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                    *//
//***********************************************************************//




#ifndef  DIST_FUNC_H
#define  DIST_FUNC_H


#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <deque>
#include <map>
#include <queue>
#include <iostream>
using namespace std;

//#include "RankedL.h"
#include "parameters.h"

// general defines for the whole system
#define NUMSIDES 2*NUMDIMS

typedef float RectReal;
typedef float PointReal;

struct Rect {
        RectReal boundary[NUMSIDES]; /* xmin,ymin,...,xmax,ymax,... */
};

struct Point {
        PointReal position[NUMDIMS]; /* xmin,ymin,...,xmax,ymax,... */
};


// ------------------------------------------------------------
// added by miki, needed datatypes
struct appDataItem { // what the application uses, a recangle plus id.
  int    the_id;
  struct Rect rect;
};
typedef map<int,appDataItem,less<int> > AnswerList; // for NN queries

struct join_pair_result { // for join queries: two appitems + their distance
  float       distance; // the distance between the two items
  appDataItem first, second; // the two items
};
typedef multimap<float,join_pair_result,less<float> > join_pair_result_list;

// ------------------------------------------------------------



class DistanceFunction {
public:
  // for multipoint query:
  multimap<float,struct Rect,less<float> > query_point_queue;
  //  in this map, the key (.first) is a float which is a
  //  WEIGHT for the point.
  float total_point_wt_sum; // adds the key from above for easy normalization
  float dimension_weights[NUMDIMS]; // a weight for each matching dimension. Must add up to 1.

  DistanceFunction(); // constructor, initialize weights
  void  MakeDefaultFunction();
  void  NormalizeWeights(); // make them add up to NUMDIMS;
  bool  add_query_point(struct Rect * query_point, float point_wt = 1.0); // add a query point
  float rectangle_mindist(struct Rect * r1, struct Rect * r2);
  float compute_distance_aux(struct Rect * query, struct Rect * item);
  float compute_distance(struct Rect * query, struct Rect * item);
  float compute_mindist_aux(struct Rect * query, struct Rect * item);
  float compute_mindist(struct Rect * query, struct Rect * item);

  // feedback specific info collection
  // .first (float) is the relevance value, .second (join_pair_result) is
  // the two matching rectangles (resp. points).
  join_pair_result_list feedback_join_data; // .first = relevance, .second = data
  multimap<float,struct Rect,less<float> > feedback_nn_data; // .first = relevance, .second = feedback
  bool feedback_start();
  int  feedback_add_tuple(float relevance, struct Rect * query, struct Rect * item);
  bool feedback_do_it();
}; // end class distance function

#endif
