//***********************************************************************//
//*        SPLIT_Q.CPP --- Define the rule to split nodes in R-Tree     *//
//*                                                                     *//
//*Copyright (C) Michael Ortega-Binderberger (miki@ics.uci.edu)         *//
//*Univ of California,Irvine Department of Information&Computer Science *//
//*                                                                     *//
//*This program is free software; you can redistribute itand/or         *//
//*modify it under the terms of the GNU General Public License          *//
//*as published by the Free Software Foundation; either version 2       *//
//*of the License, or (at your option) any later version.               *//
//*                                                                     *//
//*This program is distributed in the hope that it will be useful,      *//
//*but WITHOUT ANY WARRANTY; without even the implied warranty of       *//
//*MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *//
//*GNU General Public License for more details.                         *//
//*                                                                     *//
//*You should have received a copy of the GNU General Public License    *//
//*along with this program; if not, write to the Free Software          *//
//*Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            *//
//*02111-1307, USA.                                                     *//
//*                                                                     *//
//*                                                                     *//
//*This code was licensed under the GNU public license. For our         *//
//*modifications,the Regents of the University of California demand the *//
//*following terms: (where a conflic occurs, the lesser restrictive term*//
//*applies)                                                             *//
//*                                                                     *//
//*Copyright 1996-99 The Regents of the University of California.       *//
//*All Rights Reserved.                                                 *//
//*Permission to use, copy, modify, and distribute this software        *//
//*and its documentation for educational, research and non-profit       *//
//*purposes, without fee, and without a written agreement is hereby     *//
//*granted, provided that the above copyright notice, this paragraph    *//
//*and the following three paragraphs appear in all copies.             *//
//*Permission to incorporate this software into commercial products     *//
//*may be obtained by contacting the University of California.          *//
//*Chen Li Department of Information and Computer Science (ICS)         *//
//*University of California Irvine, California 92697-3425               *//
//*Phone: 949-824-9470.                                                 *//
//*                                                                     *//
//*This software program and documentation are copyrighted by           *//
//*The Regents of the University of California. The software program    *//
//*and documentation are supplied "as is", without any accompanying     *//
//*services from The Regents. The Regents do not warrant that the       *//
//*operation of the program will be uninterrupted or error-free.        *//
//*The end-user understands that the program was developed for          *//
//*research purposes and is advised not to rely exclusively on          *//
//*the program for any reason.                                          *//
//*                                                                     *//
//*IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO          *//
//*ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR              *//
//*CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT           *//
//*OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE       *//
//*UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF      *//
//*SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS     *//
//*ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           *//
//*WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  *//
//*THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE      *//
//*UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,  *//
//*SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                    *//
//***********************************************************************//




#include <stdio.h>
#include "assert.h"
#include "index.h"
#include "split_q.h"



/*-----------------------------------------------------------------------------
| Load branch buffer with branches from full node plus the extra branch.
-----------------------------------------------------------------------------*/
static void RTreeGetBranches(struct Node *n, struct Branch *b)
{
	register int i;

	assert(n);
	assert(b);

	/* load the branch buffer */
	for (i=0; i<NODECARD; i++)
	{
		assert(n->branch[i].child); /* n should have every entry full */
		BranchBuf[i] = n->branch[i];
	}
	BranchBuf[NODECARD] = *b;

	/* calculate rect containing all in the set */
	CoverSplit = BranchBuf[0].rect;
	for (i=1; i<NODECARD+1; i++)
	{
		CoverSplit = RTreeCombineRect(&CoverSplit, &BranchBuf[i].rect);
	}
	CoverSplitArea = RTreeRectSphericalVolume(&CoverSplit);

	RTreeInitNode(n);
}




/*-----------------------------------------------------------------------------
| Put a branch in one of the groups.
-----------------------------------------------------------------------------*/
static void RTreeClassify(int i, int group, struct PartitionVars *p)
{
	assert(p);
	assert(!p->taken[i]);

	p->partition[i] = group;
	p->taken[i] = TRUE;

	if (p->count[group] == 0)
		p->cover[group] = BranchBuf[i].rect;
	else
		p->cover[group] =
			RTreeCombineRect(&BranchBuf[i].rect, &p->cover[group]);
	p->area[group] = RTreeRectSphericalVolume(&p->cover[group]);
	p->count[group]++;
}




/*-----------------------------------------------------------------------------
| Pick two rects from set to be the first elements of the two groups.
| Pick the two that waste the most area if covered by a single rectangle.
-----------------------------------------------------------------------------*/
static void RTreePickSeeds(struct PartitionVars *p)
{
	register int i, j, seed0, seed1;
	RectReal worst, waste, area[NODECARD+1];

	for (i=0; i<NODECARD+1; i++)
		area[i] = RTreeRectSphericalVolume(&BranchBuf[i].rect);

	worst = -CoverSplitArea - 1;
	for (i=0; i<NODECARD; i++)
	{
		for (j=i+1; j<NODECARD+1; j++)
		{
			struct Rect one_rect = RTreeCombineRect(
						&BranchBuf[i].rect,
						&BranchBuf[j].rect);
			waste = RTreeRectSphericalVolume(&one_rect) -
					area[i] - area[j];
			if (waste > worst)
			{
				worst = waste;
				seed0 = i;
				seed1 = j;
			}
		}
	}
	RTreeClassify(seed0, 0, p);
	RTreeClassify(seed1, 1, p);
}




/*-----------------------------------------------------------------------------
| Copy branches from the buffer into two nodes according to the partition.
-----------------------------------------------------------------------------*/
static void RTreeLoadNodes(struct Node *n, struct Node *q,
			struct PartitionVars *p)
{
	register int i;
	assert(n);
	assert(q);
	assert(p);

	for (i=0; i<NODECARD+1; i++)
	{
		assert(p->partition[i] == 0 || p->partition[i] == 1);
		if (p->partition[i] == 0)
			RTreeAddBranch(&BranchBuf[i], n, NULL);
		else if (p->partition[i] == 1)
			RTreeAddBranch(&BranchBuf[i], q, NULL);
	}
}




/*-----------------------------------------------------------------------------
| Initialize a PartitionVars structure.
-----------------------------------------------------------------------------*/
static void RTreeInitPVars(struct PartitionVars *p)
{
	register int i;
	assert(p);

	p->count[0] = p->count[1] = 0;
	p->cover[0] = p->cover[1] = RTreeNullRect();
	p->area[0] = p->area[1] = (RectReal)0;
	for (i=0; i<NODECARD+1; i++)
	{
		p->taken[i] = FALSE;
		p->partition[i] = -1;
	}
}




/*-----------------------------------------------------------------------------
| Print out data for a partition from PartitionVars struct.
-----------------------------------------------------------------------------*/
static void RTreePrintPVars(struct PartitionVars *p)
{
	register int i;
	assert(p);

	printf("\npartition:\n");
	for (i=0; i<NODECARD+1; i++)
	{
		printf("%3d\t", i);
	}
	printf("\n");
	for (i=0; i<NODECARD+1; i++)
	{
		if (p->taken[i])
			printf("  t\t");
		else
			printf("\t");
	}
	printf("\n");
	for (i=0; i<NODECARD+1; i++)
	{
		printf("%3d\t", p->partition[i]);
	}
	printf("\n");

	printf("count[0] = %d  area = %f\n", p->count[0], p->area[0]);
	printf("count[1] = %d  area = %f\n", p->count[1], p->area[1]);
	if (p->area[0] + p->area[1] > 0)
	{
		printf("total area = %f  effectiveness = %3.2f\n",
			p->area[0] + p->area[1],
			(float)CoverSplitArea / (p->area[0] + p->area[1]));
	}
	printf("cover[0]:\n");
	RTreePrintRect(&p->cover[0], 0);

	printf("cover[1]:\n");
	RTreePrintRect(&p->cover[1], 0);
}



/*-----------------------------------------------------------------------------
| Method #0 for choosing a partition:
| As the seeds for the two groups, pick the two rects that would waste the
| most area if covered by a single rectangle, i.e. evidently the worst pair
| to have in the same group.
| Of the remaining, one at a time is chosen to be put in one of the two groups.
| The one chosen is the one with the greatest difference in area expansion
| depending on which group - the rect most strongly attracted to one group
| and repelled from the other.
| If one group gets too full (more would force other group to violate min
| fill requirement) then other group gets the rest.
| These last are the ones that can go in either group most easily.
-----------------------------------------------------------------------------*/
static void RTreeMethodZero(struct PartitionVars *p)
{
	register int i;
	RectReal biggestDiff;
	register int group, chosen, betterGroup;
	assert(p);

	RTreeInitPVars(p);
	RTreePickSeeds(p);

	while (p->count[0] + p->count[1] < NODECARD + 1
		&& p->count[0] < NODECARD + 1 - MinFill
		&& p->count[1] < NODECARD + 1 - MinFill)
	{
		biggestDiff = (RectReal)-1.;
		for (i=0; i<NODECARD+1; i++)
		{
			if (!p->taken[i])
			{
				struct Rect *r, rect_0, rect_1;
				RectReal growth0, growth1, diff;

				r = &BranchBuf[i].rect;
				rect_0 = RTreeCombineRect(r, &p->cover[0]);
				rect_1 = RTreeCombineRect(r, &p->cover[1]);
				growth0 = RTreeRectSphericalVolume(
						&rect_0)-p->area[0];
				growth1 = RTreeRectSphericalVolume(
						&rect_1)-p->area[1];
				diff = growth1 - growth0;
				if (diff >= 0)
					group = 0;
				else
				{
					group = 1;
					diff = -diff;
				}

				if (diff > biggestDiff)
				{
					biggestDiff = diff;
					chosen = i;
					betterGroup = group;
				}
				else if (diff==biggestDiff &&
					 p->count[group]<p->count[betterGroup])
				{
					chosen = i;
					betterGroup = group;
				}
			}
		}
		RTreeClassify(chosen, betterGroup, p);
	}

	/* if one group too full, put remaining rects in the other */
	if (p->count[0] + p->count[1] < NODECARD + 1)
	{
		if (p->count[0] >= NODECARD + 1 - MinFill)
			group = 1;
		else
			group = 0;
		for (i=0; i<NODECARD+1; i++)
		{
			if (!p->taken[i])
				RTreeClassify(i, group, p);
		}
	}

	assert(p->count[0] + p->count[1] == NODECARD+1);
	assert(p->count[0] >= MinFill && p->count[1] >= MinFill);
}


/*-----------------------------------------------------------------------------
| Split a node.
| Divides the nodes branches and the extra one between two nodes.
| Old node is one of the new ones, and one really new one is created.
| Tries more than one method for choosing a partition, uses best result.
-----------------------------------------------------------------------------*/
extern void RTreeSplitNode(struct Node *n, struct Branch *b, struct Node **nn)
{
	register struct PartitionVars *p;
	register int level;

	assert(n);
	assert(b);

	/* load all the branches into a buffer, initialize old node */
	level = n->level;
	RTreeGetBranches(n, b);

	/* find partition */
	p = &Partitions[0];
	RTreeMethodZero(p);

	/*
	 * put branches from buffer into 2 nodes
	 * according to chosen partition
	 */
	*nn = RTreeNewNode();
	(*nn)->level = n->level = level;
	RTreeLoadNodes(n, *nn, p);
	assert(n->count+(*nn)->count == NODECARD+1);
}
