//***********************************************************************//
//*        TEST.CPP --- main() function and conduct query               *//
//*                                                                     *//
//*Copyright (C) Michael Ortega-Binderberger (miki@ics.uci.edu)         *//
//*              Liang Jin (liangj@ics.uci.edu)                         *//
//*              Chen Li (chenli@ics.uci.edu)                           *//
//*Univ of California,Irvine Department of Information&Computer Science *//
//*                                                                     *//
//*This program is free software; you can redistribute itand/or         *//
//*modify it under the terms of the GNU General Public License          *//
//*as published by the Free Software Foundation; either version 2       *//
//*of the License, or (at your option) any later version.               *//
//*                                                                     *//
//*This program is distributed in the hope that it will be useful,      *//
//*but WITHOUT ANY WARRANTY; without even the implied warranty of       *//
//*MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        *//
//*GNU General Public License for more details.                         *//
//*                                                                     *//
//*You should have received a copy of the GNU General Public License    *//
//*along with this program; if not, write to the Free Software          *//
//*Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA            *//
//*02111-1307, USA.                                                     *//
//*                                                                     *//
//*                                                                     *//
//*This code was licensed under the GNU public license. For our         *//
//*modifications,the Regents of the University of California demand the *//
//*following terms: (where a conflic occurs, the lesser restrictive term*//
//*applies)                                                             *//
//*                                                                     *//
//*Copyright 1996-99 The Regents of the University of California.       *//
//*All Rights Reserved.                                                 *//
//*Permission to use, copy, modify, and distribute this software        *//
//*and its documentation for educational, research and non-profit       *//
//*purposes, without fee, and without a written agreement is hereby     *//
//*granted, provided that the above copyright notice, this paragraph    *//
//*and the following three paragraphs appear in all copies.             *//
//*Permission to incorporate this software into commercial products     *//
//*may be obtained by contacting the University of California.          *//
//*Chen Li Department of Information and Computer Science (ICS)         *//
//*University of California Irvine, California 92697-3425               *//
//*Phone: 949-824-9470.                                                 *//
//*                                                                     *//
//*This software program and documentation are copyrighted by           *//
//*The Regents of the University of California. The software program    *//
//*and documentation are supplied "as is", without any accompanying     *//
//*services from The Regents. The Regents do not warrant that the       *//
//*operation of the program will be uninterrupted or error-free.        *//
//*The end-user understands that the program was developed for          *//
//*research purposes and is advised not to rely exclusively on          *//
//*the program for any reason.                                          *//
//*                                                                     *//
//*IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO          *//
//*ANY PARTY FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR              *//
//*CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS, ARISING OUT           *//
//*OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE       *//
//*UNIVERSITY OF CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF      *//
//*SUCH DAMAGE. THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS     *//
//*ANY WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           *//
//*WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  *//
//*THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE      *//
//*UNIVERSITY OF CALIFORNIA HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE,  *//
//*SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.                    *//
//***********************************************************************//


#include <fstream>
#include <iostream>
#include "distance.h"

#include "index.h"
#include "fastmap.h"
using namespace std;

// ------------------------------------------------------------
//  LOADING DATA

Rtree * load_r_tree(bool load_old_tree, Rtree * &r,
                    char * file, char * persistent_file) {
  Rect rect;
  int  id;

  if (load_old_tree) {
    r->RTreeLoad(persistent_file);
  } else { // build a new one and make it persistent
    ifstream inf(file);
    if (!inf) {
      cout << "Error, could not open file: " << file << endl;
      exit(1);
    }
    while (!inf.eof()) {
      inf >> id;
      for (int i=0; i<NUMSIDES; i++) {
        inf >> rect.boundary[i];
      }
      if (!inf.eof())
        r->RTreeInsertRect(&rect, id);
    }
    r->RTreeDump(persistent_file); // now, make the tree persistent
    delete r;
    r = new Rtree;
    r->RTreeLoad(persistent_file);
  }
  return r; // r may change !!!
  // got a tree here.
}

// ------------------------------------------------------------
// ------------------------------------------------------------

int query(Rtree * rt1, Rtree * rt2, char** name, float delta) {

  DistanceFunction df;
  join_pair_result join_result;
  // parameters for collecting performance info:
  long  disk_acc1[10], disk_acc2[10], cpu_comps[10], total_cpus;
  float distance = 0.0;
  int   result=1;

  //float delta = DELTA;  // maximum mapped distance between pairs
  // open the join
  rt1->SearchJoinOpen(rt2, &df, &delta, disk_acc1, 10, disk_acc2, 10, cpu_comps, 10, & total_cpus);

  Distance1 dis;
  int falsepositive=0;

  ofstream out;
  out.open(RESULTFILE);


 result = rt1->SearchJoinNext(distance, join_result);

  while (result == 1) {

	int i=join_result.first.the_id;
    int j=join_result.second.the_id;
    if(i==1000000) i=0;
    int edit=dis.LD(name[i],name[j]);
    if(edit<=REALDELTA)
    {

		//cout << name[i] << endl;
		//cout << name[j] << endl << endl;

		out << name[i] << endl;
		out << name[j] << endl << endl;
	}
    else
        falsepositive++;

    result = rt1->SearchJoinNext(distance, join_result);

  }
  out.close();

  rt1->SearchJoinClose(); // cleanup


  /*out.open(ERRORFILE);
  out<<falsepositive;
  out.close();*/

  return 0;
}


int main(int argc, char *argv[]) {

  int i, j, str1, str2;

  int load_old_tree = false; // false= build new tree, true=use old tree
  cout << "Usage: " << argv[0] << " [old]" << endl;

  if (argc==2 && 0==strcmp(argv[1],"old")) // use old tree
    load_old_tree = true;

  char* names[SIZE];

  ifstream infile1, infile2;
  infile1.open(STR_SRC1);
  for(i=0;i<SIZE1;i++) {
      names[i]=new char[MAXLEN];
      infile1.getline(names[i],MAXLEN);
      //cout << i << names[i] << endl;
  }
  infile1.close();

  infile2.open(STR_SRC2);
  for(   ;i<SIZE;i++){
      names[i]=new char[MAXLEN];
      infile2.getline(names[i],MAXLEN);
      //cout << i << names[i] << endl;
  }
  infile2.close();


  float delta_prime;

  if(load_old_tree == false)
  {
          cout << "begin StringMap!" << endl;
          //system("date;");
          Fastmap * fm = new Fastmap(names);
          fm -> run();
          cout << "end StringMap!" << endl;
          //system("date;");

		  cout << "Estimating new delta... " << endl;
          delta_prime = fm->getDeltaPrime();
          cout << "New delta will be " << delta_prime << endl;

          if(fm != NULL) delete fm;
  }



  Rtree          *rt1 = new Rtree;
  Rtree          *rt2 = new Rtree;

  cout << "RTree fanout: " << NODECARD << " for " << NUMDIMS
       << " dimensions"<< endl;

  cout << "------- Init done. -------------------------------" << endl;
  //system("date;");

  rt1 = load_r_tree  (load_old_tree, rt1, DATASET1_SRC, DATASET1_RTREE_BIN);
  rt2 = load_r_tree  (load_old_tree, rt2, DATASET2_SRC, DATASET2_RTREE_BIN);
  cout << "------- Loading done. ----------------------------" << endl;

  cout << "==================================================" << endl;
  query(rt1, rt2, names, delta_prime);

  cout << "Outputing results to " << RESULTFILE << endl;

  cout << "==================================================" << endl;
  //system("date;");

  cout << "------- Moping up. -------------------------------" << endl;
  rt1->RTreeDump("null1"); // now, dump also deallocates.
  rt2->RTreeDump("null2"); // now, dump also deallocates.
  delete rt1;  delete rt2;

  for(int i=0; i<SIZE; i++)
      delete names[i];


  cout << "------- Done, exiting. ---------------------------" << endl;
  return 0;
}


